//Zephyr Libraries
#include "../includes/connectionManager.h"
#include <zephyr/logging/log.h>
#include <zephyr/fs/fs.h>
#include <zephyr/net/sntp.h>
//Modem Libraries
#include <modem/nrf_modem_lib.h>
#include <modem/lte_lc.h>
#include <modem/modem_info.h>
#include <zephyr/net/sntp.h>
#include <net/rest_client.h>

#include "../includes/boardDefine.h"

//Temporary settings defines
#define SNTP_IP "129.6.15.26"

//Async handler forwards
void lte_handler(const struct lte_lc_evt *const evt);
static void azure_event_handler(struct azure_iot_hub_evt *const evt);

//State variables
struct _NetworkState _NET_internalState = {
    .networkActive = false,
    .networkInited = false,
    .azureConnected = false,
    .nrfCloudConnected = false
};
struct _NetworkState *NetworkState = &_NET_internalState;

//Defualt Configuration for Thingsboard REST API
char rest_receive_buffer[RESPONSE_BUFFER_SIZE];
const char* headerFields[] = {
    "Content-Type:application/json\r\n",
    NULL
};
struct rest_client_req_context defualtRestConfiguration;


/**
 * @brief Initilizes connection through the link control library to bring up the modem
 * 
 * @return int Sucess Indicator, 0 indicates sucesful while 1 indicates that the modem was already initilized
 *              other values are error indicatorsRESPONSE_BUFFER_SIZE
 */
int Networkconnect(){
    if(_NET_internalState.networkActive == true){
        return 1; //Network already active
    }
    int err = 0;
    if (_NET_internalState.networkInited == false){
        nrf_modem_lib_init();
        lte_lc_init();
        err = lte_lc_connect_async(lte_handler);
        _NET_internalState.networkInited = true;
    }
    else{
        err = lte_lc_connect_async(lte_handler);
    }
    return err;
}

/**
 * @brief Provides event handling for the async link control connection function. This will continue
 *        to be caled upon all new events that the modem encounters. Note that some event such as the sleep
 *        one require enabling through the use of a KConfig option
 *        https://developer.nordicsemi.com/nRF_Connect_SDK/doc/latest/nrf/libraries/modem/lte_lc.html
 * 
 * @param evt Event passed into the handler by the asycn connection handler
 */
void lte_handler(const struct lte_lc_evt *const evt){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    switch (evt->type){

    case LTE_LC_EVT_NW_REG_STATUS:
        if (evt->nw_reg_status == LTE_LC_NW_REG_REGISTERED_HOME ||
            evt->nw_reg_status == LTE_LC_NW_REG_REGISTERED_ROAMING){
                LOG_INF("Network connected!");
                _NET_internalState.networkActive = true;
            }
        if (evt->nw_reg_status == LTE_LC_NW_REG_REGISTRATION_DENIED ||
                                    evt->nw_reg_status == LTE_LC_NW_REG_UNKNOWN ||
                                    evt->nw_reg_status == LTE_LC_NW_REG_UICC_FAIL)                        
            {
                LOG_WRN("Error with connection to the network!");
                _NET_internalState.networkActive = false;
            }
        break;
    
    case LTE_LC_EVT_PSM_UPDATE:
    case LTE_LC_EVT_EDRX_UPDATE:
    case LTE_LC_EVT_RRC_UPDATE:
    case LTE_LC_EVT_CELL_UPDATE:
    case LTE_LC_EVT_LTE_MODE_UPDATE:
    case LTE_LC_EVT_TAU_PRE_WARNING:
    case LTE_LC_EVT_NEIGHBOR_CELL_MEAS:
    case LTE_LC_EVT_MODEM_SLEEP_EXIT_PRE_WARNING:
    case LTE_LC_EVT_MODEM_SLEEP_EXIT:
    case LTE_LC_EVT_MODEM_SLEEP_ENTER:
    case LTE_LC_EVT_MODEM_EVENT:
            break;
    
    default:
        break;
    };
}

/**
 * @brief Sets the boards I2C RTC clock from a network time server. Network must be active before calling this
 * 
 */
void NetworksyncRTCClockSNTP(){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Polling network time server!");
    if (_NET_internalState.networkActive != true){
        LOG_ERR("NETWORK Not ready can't sync the clock");
    }
    uint32_t timeout = 1000;
    struct sntp_time buffer;
    sntp_simple(SNTP_IP, timeout, &buffer);
    RTCsetTime(&buffer.seconds);
    LOG_INF("Set time to %lln", &buffer.seconds);
}

//Initilizes connections to cloud providers
void Network_initCloud(){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Setting connection to cloud");
    if(!_NET_internalState.networkActive){
       LOG_INF("Network not active!");
    }
    //Azure Connection
    azure_iot_hub_init(azure_event_handler);
    azure_iot_hub_connect(NULL);
    //Init the Thingverse cloud
    rest_client_request_defaults_set(&defualtRestConfiguration);
    defualtRestConfiguration.resp_buff = rest_receive_buffer;
    defualtRestConfiguration.resp_buff_len = RESPONSE_BUFFER_SIZE;
    defualtRestConfiguration.http_method = HTTP_POST;
    defualtRestConfiguration.port = THINGSBOARD_PORT;
    defualtRestConfiguration.header_fields = &headerFields;
    defualtRestConfiguration.url = THINGSBOARD_URL;
    defualtRestConfiguration.host = THINGSBOARD_POST_ADDRESS;
}

void Network_ConnectToCloud(){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    if(_NET_internalState.azureConnected == true){
        LOG_INF("Already connected. Nothing to do!");
    }
    else{
        int returnValue = azure_iot_hub_connect(NULL);
        if (returnValue != 0){
            LOG_WRN("ERROR while connecting to the azure cloud, connecting fucntion returned %d", returnValue);
        }
    }
}

/**
 * @brief Event handler for the Azure initilization function
 * 
 * @param evt Passed in event
 */
static void azure_event_handler(struct azure_iot_hub_evt *const evt){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    switch (evt->type){
        
        case AZURE_IOT_HUB_EVT_CONNECTED:
            LOG_INF("Connect to Azure Cloud");
            _NET_internalState.azureConnected = true;
            break;
        case AZURE_IOT_HUB_EVT_DISCONNECTED:
            LOG_WRN("Azure Cloud disconnected!");
            _NET_internalState.azureConnected = false;
            break;
        case AZURE_IOT_HUB_EVT_CONNECTION_FAILED:
            LOG_WRN("Azure CLoud Connection failed!");
            _NET_internalState.azureConnected = false;
            break;
        default:
            break;
    }
}

void Network_MeasurementToCloud(char *JsonString){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Preparing to send measurments to cloud!");
    if (!_NET_internalState.networkActive || !_NET_internalState.azureConnected){
        LOG_ERR("NETWORK or AZURE not connected! Please connect");
        return;
    }
    struct azure_iot_hub_msg msg = {
        .topic.type = AZURE_IOT_HUB_TOPIC_EVENT,
        .payload.ptr = JsonString,
        .qos = MQTT_QOS_0_AT_MOST_ONCE,
        .payload.size = strlen(JsonString)
    };
    int err = azure_iot_hub_send(&msg);
    if (err){
        LOG_ERR("Issue sending measurements to cloud, NOT SENT");
        return;
    }
    LOG_INF("Measurements sent");
}

//Stuff for ThingsBoard

void Network_sendTo_Thingsboard(char *payload){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Sending REST Post to Things board with measurements!");
    struct rest_client_req_context telemeteryRequest = defualtRestConfiguration;
    struct rest_client_resp_context response = {};
    telemeteryRequest.body = payload;
    telemeteryRequest.body_len = strlen(payload);
    telemeteryRequest.header_fields = headerFields;
    rest_client_request(&telemeteryRequest, &response);
    LOG_INF("Response back was %s", response.http_status_code_str);
}


//Console commands
int __shell_CM_connect_to_Network(const struct shell *sh, size_t argc, char **argv){
    shell_print(sh, "Attempting to connect to network....");
    Networkconnect();
    return 0;
}

int __shell_CM_disconnect_from_Network(const struct shell *sh, size_t argc, char **argv){
    shell_print(sh, "Disconnecting from the network");
    
}

int __shell_CM_network_info(const struct shell *sh, size_t argc, char **argv){
    if (_NET_internalState.networkActive != true){
        shell_print(sh, "Network not active!");
        return 0;
    }
    modem_info_init();
    char buffer[40];
    //Get that info
    modem_info_string_get(MODEM_INFO_ICCID, buffer, sizeof(buffer));
    shell_print(sh, "ICCID: %s \n", buffer);
    modem_info_string_get(MODEM_INFO_CELLID, buffer, sizeof(buffer));
    shell_print(sh, "CELLID: %s \n", buffer);
    modem_info_string_get(MODEM_INFO_IP_ADDRESS, buffer, sizeof(buffer));
    shell_print(sh, "IP Addr: %s \n", buffer);
    modem_info_string_get(MODEM_INFO_OPERATOR, buffer, sizeof(buffer));
    shell_print(sh, "Operator: %s \n", buffer);
    return 0;
}

int __shell_CM_syncTime(const struct shell *sh, size_t argc, char **argv){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Setting RTC time from network!");
    NetworksyncRTCClockSNTP();
    return 0;
}